----------------------------------------------------------------------------
-- Barret Reducer (barret_reducer.vhd)
--
-- Computes the remainder (x mod m) using the 
-- barret algorithm
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package Barrett_reducer_package is
--  constant N: natural := 24;
  constant N: natural := 64;
  constant K: natural := 8;
  constant C: std_logic_vector(N-K downto 0) := '1'&(X"12358E75D30336"); --for N=64
--  constant C: std_logic_vector(N-K downto 0) := '1'&(X"1235"); --for N=24
  -----------------------------------------
  -- n must be greater than or equal to 2k+2
  -----------------------------------------
  constant ZERO: std_logic_vector(N - 2*K - 2 downto 0) := (others => '0');
end Barrett_reducer_package;


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.Barrett_reducer_package.all;
entity Barrett_reducer is
port (
  x: in std_logic_vector (N-1 downto 0);
  m: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (K-1 downto 0);
  done: out std_logic
);
end Barrett_reducer;

  architecture rtl of Barrett_reducer is
  signal mul_out: std_logic_vector(2*(N-K) + 1 downto 0);
  signal prod: std_logic_vector(N+2 downto 0);
  signal q, r: std_logic_vector(K+1 downto 0);
  signal dif: std_logic_vector(K+2 downto 0);
  signal mul1, mul2: std_logic_vector(N-K downto 0);
  signal sub1, sub2: std_logic_vector(K + 1 downto 0);
  signal sel_mul, sel_sub, ce_prod, ce_r, sign: std_logic;

  type states is range 0 to 6;
  signal current_state: states;

begin

  with sel_mul select mul1 <= x(N-1 downto K-1) when '0', (ZERO & q) when others;
  with sel_mul select mul2 <= c when '0', ("00" & ZERO & m) when others;
  mul_out <= mul1 * mul2;

  register_prod: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_prod = '1' then 
      prod <= mul_out(n+2 downto 0);
    end if;
  end if;
  end process;

  q <= prod(n+2 downto n-k+1);

  with sel_sub select sub1 <= x(k+1 downto 0) when '0', r when others;
  with sel_sub select sub2 <= prod(k+1 downto 0) when '0', ("00"&m) when others;

  dif <= ('0' & sub1) - ('0' & sub2);
  sign <= dif(k+2);

  register_r: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_r = '1' then 
      r <= dif(k+1 downto 0);
    end if;
  end if;
  end process;

  z <= r(k-1 downto 0);

  control_unit: process(clk, reset, current_state, sign)
  begin
  case current_state is
    when 0 to 1 => sel_mul <= '0'; ce_prod <= '0'; sel_sub <= '0'; ce_r <= '0'; done <= '1';
    when 2 => sel_mul <= '0'; ce_prod <= '1'; sel_sub <= '0'; ce_r <= '0'; done <= '0';
    when 3 => sel_mul <= '1'; ce_prod <= '1'; sel_sub <= '0'; ce_r <= '0'; done <= '0';
    when 4 => sel_mul <= '1'; ce_prod <= '0'; sel_sub <= '0'; ce_r <= '1'; done <= '0';
    when 5 => sel_mul <= '0'; ce_prod <= '0'; sel_sub <= '1'; done <= '0';
              if sign = '0' then ce_r <= '1'; else ce_r <= '0'; end if;
    when 6 => sel_mul <= '0'; ce_prod <= '0'; sel_sub <= '1'; done <= '0';
              if sign = '0' then ce_r <= '1'; else ce_r <= '0'; end if;
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '1' then current_state <= 1; end if;
      when 1 => if start = '0' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => current_state <= 4;
      when 4 => current_state <= 5;
      when 5 => if sign = '1' then current_state <= 0; else current_state <= 6; end if;
      when 6 => if sign = '1' then current_state <= 0; end if;
    end case;
  end if;
  end process;

end rtl;

