----------------------------------------------------------------------------
-- Non-restoring Reducer (nr_reducer.vhd)
--
-- Computes the remainder (x mod m) using the traditional 
-- non-restoring division algorithm
--
-- Defines 2 entities (precomputation_data_path and precomputation_reducer)
-- and a Package (precomputation_package)
-- The M value is should be Known in order to calculate the b_table
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package precomputation_package is
  constant N: integer := 16;
  constant K: integer := 8;
  constant S: integer := 2; --N/S
  constant LOGS: integer := 1; --log2(S)
  constant D: integer := 16; --
--  constant N: integer := 64;
--  constant K: integer := 8;
--  constant S: integer := 8; --N/S
--  constant LOGS: integer := 3; --log2(S)
--  constant D: integer := 18; --2**d > 255*(1+17+50+133+110+197+3+51);
  constant MAX_INDEX: std_logic_vector(LOGS-1 downto 0) := conv_std_logic_vector(s-1, LOGS);

  type digit_vector is array (0 to S-1) of std_logic_vector(K-1 downto 0);
  constant b_table: digit_vector  :=
     (conv_std_logic_vector(1,K),   conv_std_logic_vector(17,K) );
--     (conv_std_logic_vector(1,K),   conv_std_logic_vector(17,K),
--      conv_std_logic_vector(50,K),  conv_std_logic_vector(133,K),
--      conv_std_logic_vector(110,K), conv_std_logic_vector(197,K),
--      conv_std_logic_vector(3,K),   conv_std_logic_vector(51,K) );
  constant ZERO: std_logic_vector(N-D-1 downto 0) := (others => '0');
  constant SHORT_ZERO: std_logic_vector(D-K-1 downto 0) := (others => '0');
  constant m: std_logic_vector(7 downto 0) := conv_std_logic_vector(239,8);
end precomputation_package;

-----------------------------------
-- precomputation_data_path
-----------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.precomputation_package.all;
entity precomputation_data_path is
port (
  x: in std_logic_vector(N-1 downto 0);
  clk, load, reload, ce_acc, reset: in std_logic;
  sel: in std_logic_vector(LOGS-1 downto 0);
  end_of_computation: out std_logic;
  z: out std_logic_vector(K-1 downto 0)
);
end precomputation_data_path;

architecture rtl of precomputation_data_path is
signal r: std_logic_vector(N-1 downto 0);
signal vector_r: digit_vector;
signal vector_i, b_i: std_logic_vector(K-1 downto 0);
signal product: std_logic_vector(2*K-1 downto 0);
signal next_acc, acc: std_logic_vector(D-1 downto 0);
signal r_minus_m: std_logic_vector(k downto 0);

begin

  digit_selection: for i in 0 to s-1 generate
    vector_r(i) <= r((i+1)*K-1 downto i*K);
  end generate;
  vector_i <= vector_r(conv_integer(sel));
  b_i <= b_table(conv_integer(sel));
  product <= vector_i * b_i;
  next_acc <= acc + product;

  register_acc: process(reset, clk)
  begin
    if reset = '1' then acc <= (others => '0');
    elsif clk'event and clk = '1' then
      if load = '1' or reload = '1' then acc <= (others => '0');
      elsif ce_acc = '1' then acc <= next_acc;
      end if;
    end if;
  end process register_acc;

  register_r: process(reset, clk)
  begin
    if reset = '1' then r <= (others => '0');
    elsif clk'event and clk = '1' then
      if load = '1' then r <= x;
      elsif reload = '1' then r <= ZERO & acc;
      end if;
    end if;
  end process register_r;

  end_of_computation <= '1' when r(D-1 downto K) = SHORT_ZERO else '0';
  r_minus_m <= ('0' & r(K-1 downto 0)) - ('0' & m);
  with r_minus_m(K) select z <= r(K-1 downto 0) when '1', r_minus_m(K-1 downto 0) when others;

end rtl;

-----------------------------------
-- precomputation_reducer
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.precomputation_package.all;
entity precomputation_reducer is
port (
  x: in std_logic_vector (N-1 downto 0);
  m: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (K-1 downto 0);
  done: out std_logic
);
end precomputation_reducer;

architecture rtl of precomputation_reducer is

component precomputation_data_path is
port (
  x: in std_logic_vector(N-1 downto 0);
  clk, load, reload, ce_acc, reset: in std_logic;
  sel: in std_logic_vector(LOGS-1 downto 0);
  end_of_computation: out std_logic;
  z: out std_logic_vector(K-1 downto 0)
);
end component;

signal load, reload, ce_acc, end_of_computation: std_logic;
signal sel: std_logic_vector(LOGS-1 downto 0);
type states is range 0 to 5;
signal current_state: states;

begin

data_path: precomputation_data_path port map 
  (x, clk, load, reload, ce_acc, reset, sel, end_of_computation, z);

counter: process(reset, clk)
begin
  if reset = '1' then sel <= (others => '0');
  elsif clk' event and clk = '1' then
    if load = '1' or reload = '1' then sel <= (others => '0');
    elsif ce_acc = '1' then sel <= sel+1; 
  end if;
  end if;
end process counter;

control_unit: process(clk, reset, current_state)
begin
  case current_state is
    when 0 to 1 => load <= '0'; reload <= '0'; ce_acc <= '0'; done <= '1';
    when 2 => load <= '1'; reload <= '0'; ce_acc <= '0'; done <= '0';
    when 3 => load <= '0'; reload <= '0'; ce_acc <= '1'; done <= '0';
    when 4 => load <= '0'; reload <= '1'; ce_acc <= '0'; done <= '0';
    when 5 => load <= '0'; reload <= '0'; ce_acc <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
    when 0 => if start = '1' then current_state <= 1; end if;
    when 1 => if start = '0' then current_state <= 2; end if;
    when 2 => current_state <= 3;
    when 3 => if sel = MAX_INDEX then current_state <= 4; end if;
    when 4 => current_state <= 5;
    when 5 => if end_of_computation = '1' then current_state <= 0;
              else current_state <= 3; end if;
    end case;
  end if;
end process control_unit;

end rtl;
